@@ -1,22 +1,25 @@
 # coding: utf-8
 
-from __future__ import print_function, absolute_import, division, unicode_literals
-
 import glob
 import os
 import sys
 import inspect
 import subprocess
+import argparse
 import datetime
 from textwrap import dedent
 from ruamel.std.pathlib import Path
+from ruamel.std.pathlib.tempdir import TempDir
 from ruamel.yaml import YAML
-from ruamel.yaml.scalarstring import PreservedScalarString
+from ruamel.yaml.scalarstring import LiteralScalarString
+
+from typing import Optional, Any, Union, Dict, Set, List, TextIO
+
+from .stdout import StdOut
 
 Bugs = """\
 - check empty documents. E.g. !python that only has code in !pre, and !stdout
   without any introductory text
-- running with python2 on ruamel.yaml doc will raise str vs unicode issue
 """
 
 ToDo = """\
@@ -35,7 +38,7 @@ ToDo = """\
 - support code-block directive  http://rst2pdf.ralsina.me/handbook.html#syntax-highlighting
 - list structure of the .ryd file
 - process documents using xyz:prog  matching the document tag xyz by rt piping through prog
-- why do toplevel PreservedScalarString dumps have |2, it is not even correct!
+- why do toplevel LiteralScalarString dumps have |2, it is not even correct!
 = consider moving to plugin system
 """
 
@@ -45,13 +48,13 @@ class RydExecError(Exception):
 
 
 class RYD(object):
-    def __init__(self, args, config):
+    def __init__(self, args: argparse.Namespace, config: Optional[Any] = None) -> None:
         self._args = args
         self._config = config
         self._name_printed = False
-        self._current_path = None
+        self._current_path: Union[Path, None] = None
 
-    def convert(self):
+    def convert(self) -> None:
         for file_name in self._args.file:
             self._name_printed = False
             if '*' in file_name or '?' in file_name:
@@ -60,11 +63,11 @@ class RYD(object):
                 continue
             self.convert_one(Path(file_name), verbose=self._args.verbose)
 
-    def clean(self):
+    def clean(self) -> None:
         for file_name in self._todo():
             self.convert_one(file_name, clean=True)
 
-    def _todo(self):
+    def _todo(self) -> list[Path]:
         todo = []
         for file_name in self._args.file:
             self._name_printed = False
@@ -80,15 +83,17 @@ class RYD(object):
         # print('todo', todo)
         return todo
 
-    def name(self):
+    def name(self) -> None:
         """print name of file only once (either verbose or on error)"""
         if self._name_printed:
             return
         self._name_printed = True
         print(self._current_path)
 
-    def convert_one(self, path, clean=False, verbose=0):
+    def convert_one(self, path: Path, clean: bool = False, verbose: int = 0) -> None:
         from .stackoverflow import StackOverflowConvertor
+        from .markdown import MarkdownConvertor
+
         if self._current_path is None and not path.exists():
             print('unknown command, or file:', path)
             sys.exit(1)
@@ -96,12 +101,14 @@ class RYD(object):
         if verbose > 0:
             self.name()
         yaml = YAML()
-        convertor = None
+        convertor: Optional[Convertor] = None
         for x in yaml.load_all(path):
             if convertor is None:
-                assert 0.099 < float(x['version']) < 0.101
+                assert 0.099 < float(x['version']) < 0.201
                 if x['output'] == 'rst':
                     convertor = RSTConvertor(self, yaml, x, path)
+                elif x['output'] == 'md':
+                    convertor = MarkdownConvertor(self, yaml, x, path)
                 elif x['output'] == 'so':
                     convertor = StackOverflowConvertor(self, yaml, x, path)
                 else:
@@ -112,7 +119,10 @@ class RYD(object):
                 return
             if not convertor(x):  # already up-to-date
                 sys.stdout.flush()
+                # ToDo you cannot return here, the PDF might not be generated because
+                # of some rst2pdf issue
                 return
+        assert convertor is not None
         if convertor.updated:
             if verbose > 0:
                 print('updated')
@@ -120,7 +130,7 @@ class RYD(object):
         # convertor.dump()
         sys.stdout.flush()
 
-    def from_rst(self):
+    def from_rst(self) -> None:
         for file_name in (Path(f) for f in self._args.file):
             ryd_name = file_name.with_suffix('.ryd')
             if ryd_name.exists() and not self._args.force:
@@ -142,11 +152,11 @@ class RYD(object):
                 )
                 fp.write(rst.update_sections())
 
-    def roundtrip(self):
+    def roundtrip(self) -> None:
         for file_name in self._todo():
             self.round_trip_one(file_name)
 
-    def round_trip_one(self, path):
+    def round_trip_one(self, path: Path) -> None:
         self._current_path = path
         if self._args.verbose > 0:
             self.name()
@@ -157,33 +167,35 @@ class RYD(object):
                 convertor = RoundTripConvertor(self, yaml, x, path)
                 continue
             convertor(x)
+        assert convertor is not None
         if convertor.updated:
             convertor.write()
 
 
 class reStructuredText(object):
-    def __init__(self, path):
+    def __init__(self, path: Path) -> None:
         self._path = path
-        self._lines = None
+        self._lines: Union[List[str], None] = None
         self._blank = set([-1])  # in case you have a divider on the first line
-        self._section_lines = {}
-        self._divider_lines = set()
+        self._section_lines: Dict[int, str] = {}
+        self._divider_lines: Set[Any] = set()
         self._double_section = ""  # above and below
-        self._double_section_lines = {}
+        self._double_section_lines: Dict[Any, Any] = {}
         self._single_section = ""  # only below
-        self._single_section_lines = {}
+        self._single_section_lines: Dict[Any, Any] = {}
 
     @property
-    def lines(self):
+    def lines(self) -> list[str]:
         if self._lines is None:
             self._lines = self._path.read_text().splitlines()
+        assert isinstance(self._lines, list)
         return self._lines
 
-    def analyse_sections(self):
+    def analyse_sections(self) -> None:
         non_alpha = '\'"#*=-^+'
         non_alpha = '=-`:\'"~^_*+#<>.'
 
-        def char_repeats(line):
+        def char_repeats(line: str) -> bool:
             """assume a stripped line, check if the first character is repeated"""
             idx = len(line)
             while idx > 1:
@@ -255,7 +267,7 @@ class reStructuredText(object):
         print('single:', self._single_section)
         print('       ', self._single_section_lines)
 
-    def update_sections(self):
+    def update_sections(self) -> str:
         # almost the Python recomendation, but not using '-' for sections
         double = '#*'
         single = '=+^"'
@@ -275,17 +287,17 @@ class reStructuredText(object):
 ##############################################################################
 
 
-class RydDoc(PreservedScalarString):
+class RydDoc(LiteralScalarString):
 
     needs_double_colon = True
 
     @classmethod
-    def from_yaml(cls, constructor, node):
+    def from_yaml(cls, constructor: Any, node: Any) -> Any:
         return cls(node.value)
 
     @classmethod
-    def to_yaml(cls, representer, node):
-        return representer.represent_preserved_scalarstring(cls.yaml_tag, node.value)
+    def to_yaml(cls, representer: Any, node: Any) -> Any:
+        return representer.represent_literal_scalarstring(cls.yaml_tag, node.value)  # type: ignore # NOQA
 
 
 # class BashRaw(RydDoc):
@@ -389,7 +401,7 @@ class StdoutRaw(RydDoc):
 class RydProgram(RydDoc):
     needs_double_colon = True
 
-    def dump(self, fp, convertor):
+    def dump(self, fp: TextIO, convertor: 'Convertor') -> None:
         """generic include of code with indentation unless empty"""
         for line in self.strip().splitlines():
             if line.strip():
@@ -401,6 +413,9 @@ class RydProgram(RydDoc):
         # not using nim prefix when unpy-ing
         self.last_output = self.check_output(convertor)
 
+    def check_output(self, convertor: 'Convertor') -> Union[str, None]:
+        raise NotImplementedError
+
 
 #########
 # Nim
@@ -411,11 +426,14 @@ class Nim(RydProgram):
 
     yaml_tag = '!nim'
 
-    def check_output(self, convertor):
+    def check_output(self, convertor: 'Convertor') -> Union[str, None]:
         # None is replaced by the filename
-        cmd = ["nim", "compile", "--verbosity:0", "--hint[Processing]:off", None]
-        convertor.last_compile = convertor.check_output(convertor.nim_pre + self,
-                                                        exe=cmd, ext=".nim")
+        cmd = ['nim', 'compile', '--verbosity:0', '--hint[Processing]:off', None]
+        convertor.last_compile = convertor.check_output(
+            convertor.nim_pre + self,  # type: ignore
+            exe=cmd,
+            ext='.nim',
+        )
         # print('convertor.last_compile', convertor.last_compile)
         exe = convertor.last_source.with_suffix("")
         if exe.exists():
@@ -433,8 +451,8 @@ class NimPre(RydDoc):
 
     yaml_tag = '!nim-pre'
 
-    def dump(self, fp, convertor):
-        convertor.nim_pre = self
+    def dump(self, fp: Any, convertor: 'Convertor') -> None:
+        convertor.nim_pre = self  # type: ignore
 
 
 class UnPy(RydProgram):
@@ -443,53 +461,60 @@ class UnPy(RydProgram):
 
     yaml_tag = '!unpy'
 
-    def check_output(self, convertor):
+    def check_output(self, convertor: 'Convertor') -> Union[str, None]:
         # None is replaced by the filename
-        return convertor.check_output(self, exe=["unpy", "update", "--output:-", None],
-                                      ext=".nim")
+        return convertor.check_output(
+            self, exe=['unpy', 'update', '--output:-', None], ext='.nim'
+        )
 
 
 ##############################################################################
 
 
 class Convertor(object):
-    def __init__(self, ryd):
+    def __init__(self, ryd: RYD) -> None:
         self._ryd = ryd
-        self._tempdir = None
+        self._tempdir: Optional[TempDir] = None
         self._tmpfile_nr = 0
-        self._tag_obj = {}
+        self._tag_obj: Dict[str, Any] = {}
         self._tag_doc = self.gather_tag_documentation()
+        self.last_compile: Any = None
+        self._path: Path
+        self.updated = False
+        self._out_path: Union[Path, StdOut]
 
-    def gather_tag_documentation(self):
+    def gather_tag_documentation(self) -> Dict[str, List[Any]]:
         tag_doc = {}
         for _name, obj in inspect.getmembers(sys.modules[__name__]):
-            if inspect.isclass(obj) and obj is not RydDoc and issubclass(obj, RydDoc) \
-               and hasattr(obj, 'yaml_tag'):
-                self._tag_obj[obj.yaml_tag] = obj
+            if (
+                inspect.isclass(obj)
+                and obj is not RydDoc
+                and issubclass(obj, RydDoc)
+                and hasattr(obj, 'yaml_tag')
+            ):
+                self._tag_obj[obj.yaml_tag] = obj  # type: ignore
                 d = obj.__doc__ if obj.__doc__ is not None else 'N/A\n  N/A'
                 d1, d2 = d.lstrip().split('\n', 1)
                 d1 = d1.strip()
                 d2 = dedent(d2).lstrip()
-                tag_doc[obj.yaml_tag] = [d1, d2]
+                tag_doc[obj.yaml_tag] = [d1, d2]  # type: ignore
         # for k in sorted(self.tags):
         #     print(k, '->', self.tags[k])
         return tag_doc
 
     @property
-    def tempdir(self):
-        from ruamel.std.pathlib.tempdir import TempDir
-
+    def tempdir(self) -> TempDir:
         if self._tempdir is None:
             self._tempdir = TempDir(prefix='ryd', keep=3)
         return self._tempdir
 
-    def get_tags(self):
+    def get_tags(self) -> Dict[str, List[Any]]:
         return self._tag_doc
 
-    def update(self, s):
+    def update(self, s: Any) -> Any:
         return s
 
-    def clean(self):
+    def clean(self) -> None:
         assert self._path.exists()
         if self._out_path.exists():
             self._out_path.unlink()
@@ -497,7 +522,9 @@ class Convertor(object):
     python_exe = os.environ.get('RYD_PYTHON', sys.executable)
     # print('python_exe', python_exe, sys.executable)
 
-    def check_output(self, s, exe=None, ext=None):
+    def check_output(
+        self, s: str, exe: Optional[list[Any]] = None, ext: Optional[str] = None
+    ) -> Union[str, None]:
         if exe is None:
             exe = [self.python_exe, None]
             ext = '.py'
@@ -521,9 +548,15 @@ class Convertor(object):
                 )
             return None
 
+    def __call__(self, s: Any) -> bool:
+        raise NotImplementedError
+
+    def write(self) -> None:
+        raise NotImplementedError
+
 
 class RSTConvertor(Convertor):
-    def __init__(self, ryd, yaml, md, path):
+    def __init__(self, ryd: RYD, yaml: YAML, md: Dict[str, Any], path: Path) -> None:
         super().__init__(ryd)
         self._yaml = yaml
         for v in self._tag_obj.values():
@@ -531,19 +564,19 @@ class RSTConvertor(Convertor):
         self._md = md
         self._path = path
         self._out_path = self._path.with_suffix('.rst')
-        self.data = []
+        self.data: List[Any] = []
         ####
-        self.last_output = ""
+        self.last_output: Union[str, None] = ""
         self.updated = False
         self.python_pre = ""
         # register all defined classes and store tag + comment
 
-    def __call__(self, s):
+    def __call__(self, s: Any) -> bool:
         # the read is at the end
         if (
             not self._ryd._args.force
             and self._out_path.exists()
-            and self._path.stat().st_mtime < self._out_path.stat().st_mtime
+            and self._path.stat().st_mtime < self._out_path.stat().st_mtime  # type: ignore
         ):
             if self._ryd._args.verbose > 0:
                 print('skipping', end=' ')
@@ -562,14 +595,15 @@ class RSTConvertor(Convertor):
             self.data.append(s)
         return True
 
-    def update(self, s):
+    def update(self, s: Any) -> bool:
         # if isinstance(s, RydDoc) and not isinstance(s, Python):
         #     print(type(s))
         if isinstance(s, Python):
-            return None
+            return False
         # find all backquotes in document
         bqs = []
-        lines = [0]  # character index in file to beginning of each line
+        # character index in file to beginning of each line
+        lines: List[Union[int, None]] = [0]
         line = 0
         col = 0
         for idx, ch in enumerate(s):
@@ -588,7 +622,9 @@ class RSTConvertor(Convertor):
             lnr = bqs[bqidx][1]
             # ``pair of double backquotes`` -> code
             try:
-                if s[lines[lnr]] == ' ' and s[lines[lnr] + 1] == ' ':
+                assert isinstance(lines[lnr], int)
+                tlnr: int = lines[lnr]  # type: ignore
+                if s[tlnr] == ' ' and s[tlnr + 1] == ' ':
                     # first characters on line are spaces -> code/ryd example
                     bqidx += 1
                     continue
@@ -626,15 +662,17 @@ class RSTConvertor(Convertor):
                 last_line_displayed = lnr
             print(' ' * (-1 + len(str(lnr + self._line)) + bqs[bqidx][2]), '--^')
             bqidx += 1
-        return None
+        return False
 
-    def rst2pdf(self, file_path):
+    def rst2pdf(self, file_path: Union[Path, StdOut]) -> str:
         if self._ryd._args.verbose > 0:
             print('generating PDF')
         fn = str(file_path)
-        return subprocess.check_output(['rst2pdf', fn, '-s', 'freetype-sans,eightpoint'])
+        return subprocess.check_output(
+            ['rst2pdf', fn, '-s', 'freetype-sans,eightpoint'], encoding='utf-8'
+        )
 
-    def write(self):
+    def write(self) -> None:
         with self._out_path.open('w') as fp:
             try:
                 self.dump(fp, base_dir=self._out_path.parent)
@@ -656,6 +694,7 @@ class RSTConvertor(Convertor):
             if self._md.get('encrypt'):
                 pw = self._md['encrypt']['passwd']
                 # print('pw', pw, file=sys.stderr)
+                assert isinstance(self._out_path, Path)
                 pdf_file = self._out_path.with_suffix('.pdf')
                 enc_file = self._out_path.with_suffix(
                     '.{:%Y%m%d}.pdf'.format(datetime.date.today())
@@ -665,7 +704,7 @@ class RSTConvertor(Convertor):
                     ['pdftk', str(pdf_file), 'output', str(enc_file), 'user_pw', pw]
                 )
 
-    def dump(self, fp=sys.stdout, base_dir=None):
+    def dump(self, fp: Any = sys.stdout, base_dir: Optional[Path] = None) -> None:
         last_ended_in_double_colon = False
         last_code = False
         for d in self.data:
@@ -707,17 +746,12 @@ class RSTConvertor(Convertor):
                             print('  ', line, sep="", end="", file=fp)
                         else:
                             print('', line, sep="", end="", file=fp)
-            elif isinstance(d, PythonHidden):
-                self.check_output(self.python_pre + d)
-                self.last_output = ""
-                if self._ryd._args.verbose > 1:
-                    print('=========== output =========')
-                    print(self.last_output, end="")
-                    print('============================')
-                if self.last_output is None:
-                    raise RydExecError("error executing Python")
-            elif isinstance(d, (Python, Code)):
-                if d:
+            # elif isinstance(d, PythonHidden):
+            #    self.last_output = self.check_output(self.python_pre + d)
+            #    if self.last_output is None:
+            #        raise RydExecError("error executing Python")
+            elif isinstance(d, (Python, Code, PythonHidden)):
+                if d and not isinstance(d, PythonHidden):
                     if not last_ended_in_double_colon:
                         print('\n::\n', file=fp)
                     for line in d.strip().splitlines():
@@ -729,14 +763,14 @@ class RSTConvertor(Convertor):
                     last_code = True
                 else:
                     pass  # empty segment, might have pre-
-                if isinstance(d, Python):
+                if isinstance(d, (Python, PythonHidden)):
                     self.last_output = self.check_output(self.python_pre + d)
                     if self._ryd._args.verbose > 1:
                         print('=========== output =========')
                         print(self.last_output, end="")
                         print('============================')
                     if self.last_output is None:
-                        raise RydExecError("error executing Python")
+                        raise RydExecError('error executing Python')
             elif isinstance(d, PythonPre):
                 self.python_pre = d
             elif isinstance(d, Comment):
@@ -756,20 +790,22 @@ class RSTConvertor(Convertor):
                     prefix = "" if isinstance(d, StdoutRaw) else '  '
                     op = self.last_compile if isinstance(d, LastCompile) else self.last_output
                     if op is None:
-                        print('{}>>>>>>>>>  NO OUTPUT GENERATED <<<<<<<<<'.format(prefix),
-                              file=fp)
+                        print(
+                            '{}>>>>>>>>>  NO OUTPUT GENERATED <<<<<<<<<'.format(prefix),
+                            file=fp,
+                        )
                     else:
                         for line in op.rstrip().splitlines():
                             print('{}{}'.format(prefix, line), file=fp)
                     last_code = True
-                elif type(d) != PreservedScalarString:
+                elif type(d) != LiteralScalarString:
                     print('found unknown document type:', d.yaml_tag)
                     sys.exit(1)
                 # print(file=fp)
 
 
 class RoundTripConvertor(Convertor):
-    def __init__(self, ryd, yaml, md, path):
+    def __init__(self, ryd: RYD, yaml: YAML, md: Dict[str, Any], path: Path) -> None:
         super().__init__(ryd)
         self._yaml = yaml
         # for v in self._tag_obj.values():
@@ -781,28 +817,30 @@ class RoundTripConvertor(Convertor):
         self.last_output = ""
         self.updated = False
 
-    def __call__(self, x):
+    def __call__(self, x: Any) -> bool:
         if not self._ryd._args.oitnb:
             self.data.append(x)
-            return
+            return True
         try:
             v = x.tag.value
         except AttributeError:
             self.data.append(x)
-            return
+            return True
         if not v.startswith('!python') or not x.value.strip():
             self.data.append(x)
-            return
+            return True
         # only python with a real body
-        y = subprocess.check_output(['oitnb', '-q', '-'], input=bytes(x.value, 'utf-8'))
-        y = y.decode('utf-8')
+        y = subprocess.check_output(
+            ['oitnb', '-q', '-'], input=bytes(x.value, 'utf-8')
+        ).decode('utf-8')
         if x.value != y:
             print(x.value, y)
             self.updated = True
             x.value = y + '\n'
         self.data.append(x)
+        return True
 
-    def write(self):
+    def write(self) -> None:
         print('writing')
         self._yaml.explicit_start = True
         self._yaml.dump_all(self.data, self._out_path)
